#include <ptlib.h>
#include <math.h>
#include "PFFTSoundChannel.h"

PFFTSoundChannel::PFFTSoundChannel()
{
	CommonConstruct();
}

PFFTSoundChannel::PFFTSoundChannel(
  const PString & device,       /// Name of sound driver/device
  Directions dir,               /// Sound I/O direction
  unsigned numChannels ,     /// Number of channels eg mono/stereo
  unsigned sampleRate ,   /// Samples per second
  unsigned bitsPerSample   /// Number of bits per sample
  ) : PSoundChannel(device,dir,numChannels,sampleRate,bitsPerSample)
{
	CommonConstruct();
}


void PFFTSoundChannel::CommonConstruct(void)
{
	m_DTMFCount = 0;
	m_lastDTMF=0;
	m_lastAcceptedDTMF=0;
	m_lastOffset=0;
}


BOOL PFFTSoundChannel::Write(
  const void * buf, /// Pointer to a block of memory to write.
  PINDEX len        /// Number of bytes to write.
)
{
#if 1
	int i=0;
	// copy FFT_POINTS.  remember len MOD FFT_POINTS may not be zero, so finish off on next call
	while (i<len)
	{
		do
		{
			// convert to floating point.  Source is 16bit PCM
			m_realIn[m_lastOffset] = ((short*)buf)[i/2];
			m_lastOffset++;	
			i += sizeof(short);
		}
		while ((m_lastOffset<FFT_POINTS) && (i<len));

		if (m_lastOffset == FFT_POINTS) // got enough data to do the FFT
		{
			m_lastOffset = 0;
			fft_float(FFT_POINTS,FALSE,m_realIn,NULL,m_realOut,m_imagOut);	// do the FFT

			// For the frequency range 0-4kHz we should find two tones if DTMF is present.
			// Split the domain into 0-1kHz, and 1-4kHz.  Each array entry represents 8000/FFT_POINTS Hz
			float domf1 = FindDominantFrequency(FFT_POINTS,0,1000/(8000/FFT_POINTS),m_realOut,m_imagOut,m_ampl);
			// now 1kHz to 4kHz
			float domf2 = FindDominantFrequency(FFT_POINTS,1000/(8000/FFT_POINTS)+1,FFT_POINTS/2,m_realOut,m_imagOut,m_ampl);

			char DTMF = FindDTMF((short)domf1,(short)domf2);
			
			if (DTMF)
			{
				//printf("%c\t%f\t%f\r\n",DTMF, domf1,domf2);
				if (DTMF != m_lastDTMF)		// if we got say 25ms of '1' and 25ms of'*', reject the first
					m_DTMFCount = 0;

				m_DTMFCount++;
				if (m_DTMFCount == MIN_DTMF_COUNT) // we require at least 50ms worth (for Australia at least)
				{
					if (m_lastAcceptedDTMF != DTMF) // this is a new tone, not a continuing long tone
					{
						m_lastAcceptedDTMF = DTMF;
						printf("DTMF %c\r\n",DTMF);			// TODO - ??? somehow invoke something on H323Connection???
					}
					m_DTMFCount = 0;
				}
				m_lastDTMF = DTMF;
			}
			else	// noise/silence.  NOT DTMF.
			{
				m_DTMFCount = 0;
				m_lastDTMF = 0;
				m_lastAcceptedDTMF = 0;
			}
		}
	}
#endif

	// TODO - if we decode DTMF, then send a buffer silence to the sound card!
	return PSoundChannel::Write(buf,len);
}






/////////////////////////////////////////////////////////////////////////////
typedef struct
{
	char  tone;
	short freql;
	short freqh;
} tDTMF;

static const tDTMF DTMF[] = 
	{	{'1', 697, 1209},
		{'2', 697, 1336},
		{'3', 697, 1477},
		{'A', 697, 1633},
		{'4', 770, 1209},
		{'5', 770, 1336},
		{'6', 770, 1477},
		{'B', 770, 1633},
		{'7', 852, 1209},
		{'8', 852, 1336},
		{'9', 852, 1477},
		{'C', 852, 1633},
		{'*', 941, 1209},
		{'0', 941, 1336},
		{'#', 941, 1477},
		{'D', 941, 1633}	};

#define MAX_FREQ_ERROR	(8000/FFT_POINTS)	// +/- error in Hz, reduce this if FFT points goes up.

char PFFTSoundChannel::FindDTMF(short freql, short freqh)
{
	for (int i=0; i<16; i++)
	{
		tDTMF *xDTMF = (tDTMF*) &DTMF[i];
		if ((freql >= (xDTMF->freql - MAX_FREQ_ERROR)) && (freql <= (xDTMF->freql + MAX_FREQ_ERROR)) &&
			(freqh >= (xDTMF->freqh - MAX_FREQ_ERROR)) && (freqh <= (xDTMF->freqh + MAX_FREQ_ERROR)) )
			return xDTMF->tone;
	}
	return 0;
}





/////////////////////////////////////////////////////////////////////////////
// This function returns the dominant frequency in Hz.

#define windowSize 4	// width of search window (experiment with this; try 4)

float PFFTSoundChannel::FindDominantFrequency (
    int numSamples,           // total number of frequency samples in buffer
	int start,
	int stop,				 // number of positive frequencies to stop at
    float *real,             // real coefficient buffer
    float *imag,             // imaginary coefficient buffer
    float *ampl)             // work buffer for holding r*r+i*i values
{
    // Pre-calculate amplitude values in the buffer provided by the caller.
    // This makes the code a lot more efficient.

    for ( int i=start; i<stop; i++ )
    {
        ampl[i] = real[i]*real[i] + imag[i]*imag[i];
    }

    // Zero out amplitudes for frequencies below 20 Hz.
	if (start==0)
	{
		int low = int ( numSamples * 20.0 / 8000 );
		//ASSERT ( low>=0 && low<numSamples );
		for ( i=low; i>=0; i-- )
		{
			ampl[i] = 0.0;
		}
	}

    // In the following loop, i1 is the left side of the window
    // and i2 is the right side.

    float maxWeight = -1.0;
    float centerFreq = -1.0;

    for ( int i2 = start+windowSize; i2 < stop; i2++ )
    {
        const int i1 = i2 - windowSize;

        float moment = 0.0;
        float weight = 0.0;

        for ( int i=i1; i<=i2; i++ )
        {
            weight += ampl[i];
            moment += float(i) * ampl[i];
        }

        if ( weight > maxWeight )
        {
            maxWeight = weight;
            centerFreq = moment / weight;
        }
    }

    return 8000 * centerFreq / float(numSamples);
}





/*============================================================================

    fourierf.c  -  Don Cross <dcross@intersrv.com>

    http://www.intersrv.com/~dcross/fft.html

    Contains definitions for doing Fourier transforms
    and inverse Fourier transforms.

    This module performs operations on arrays of 'float'.

    Revision history:

1998 September 19 [Don Cross]
    Updated coding standards.
    Improved efficiency of trig calculations.

============================================================================*/
#define  DDC_PI  (3.14159265358979323846)

/////////////////////////////////////////////////////////////////////////////
unsigned PFFTSoundChannel::NumberOfBitsNeeded ( unsigned PowerOfTwo )
{
	unsigned i;
	for ( i=0; ; i++ )
	{
		if ( PowerOfTwo & (1 << i) )
			return i;
	}
}


/////////////////////////////////////////////////////////////////////////////
unsigned PFFTSoundChannel::ReverseBits ( unsigned index, unsigned NumBits )
{
    unsigned i, rev;

    for ( i=rev=0; i < NumBits; i++ )
    {
        rev = (rev << 1) | (index & 1);
        index >>= 1;
    }

    return rev;
}

/////////////////////////////////////////////////////////////////////////////
void PFFTSoundChannel::fft_float (
    unsigned  NumSamples,
    int       InverseTransform,
    float    *realIn,
    float    *ImagIn,
    float    *realOut,
    float    *imagOut )
{
    unsigned NumBits;    /* Number of bits needed to store indices */
    unsigned i, j, k, n;
    unsigned BlockSize, BlockEnd;

    double angle_numerator = 2.0 * DDC_PI;
    double tr, ti;     /* temp real, temp imaginary */


    if ( InverseTransform )
        angle_numerator = -angle_numerator;

    NumBits = NumberOfBitsNeeded ( NumSamples );

    /*
    **   Do simultaneous data copy and bit-reversal ordering into outputs...
    */

    for ( i=0; i < NumSamples; i++ )
    {
        j = ReverseBits ( i, NumBits );
        realOut[j] = realIn[i];
        imagOut[j] = (ImagIn == NULL) ? (float)0.0 : ImagIn[i];
    }

    /*
    **   Do the FFT itself...
    */

    BlockEnd = 1;
    for ( BlockSize = 2; BlockSize <= NumSamples; BlockSize <<= 1 )
    {
        double delta_angle = angle_numerator / (double)BlockSize;
        double sm2 = sin ( -2 * delta_angle );
        double sm1 = sin ( -delta_angle );
        double cm2 = cos ( -2 * delta_angle );
        double cm1 = cos ( -delta_angle );
        double w = 2 * cm1;
        double ar[3], ai[3];

        for ( i=0; i < NumSamples; i += BlockSize )
        {
            ar[2] = cm2;
            ar[1] = cm1;

            ai[2] = sm2;
            ai[1] = sm1;

            for ( j=i, n=0; n < BlockEnd; j++, n++ )
            {
                ar[0] = w*ar[1] - ar[2];
                ar[2] = ar[1];
                ar[1] = ar[0];

                ai[0] = w*ai[1] - ai[2];
                ai[2] = ai[1];
                ai[1] = ai[0];

                k = j + BlockEnd;
                tr = ar[0]*realOut[k] - ai[0]*imagOut[k];
                ti = ar[0]*imagOut[k] + ai[0]*realOut[k];

                realOut[k] = realOut[j] - (float)tr;
                imagOut[k] = imagOut[j] - (float)ti;

                realOut[j] += (float)tr;
                imagOut[j] += (float)ti;
            }
        }

        BlockEnd = BlockSize;
    }

    /*
    **   Need to normalize if inverse transform...
    */

    if ( InverseTransform )
    {
        double denom = (double)NumSamples;

        for ( i=0; i < NumSamples; i++ )
        {
            realOut[i] /= (float)denom;
            imagOut[i] /= (float)denom;
        }
    }
}

