/*
 * xlibvid.cxx                  
 *                            
 *This file contains the class heirachy for 1)creating windows under xlib.
 *                                          2)displaying data on those windows
 *                                          3)converting image formats.
 *
 * Copyright (c) 1999-2000 Indranet Technologies ltd
 * 
 * The contents of this file are subject to the Mozilla Public License
 * Version 1.0 (the "License"); you may not use this file except in
 * compliance with the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 * 
 * Software distributed under the License is distributed on an "AS IS"
 * basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
 * the License for the specific language governing rights and limitations
 * under the License.
 * 
 * The Original Code is Open H323
 * 
 * The Initial Developer of the Original Code is Equivalence Pty. Ltd. 
 * 
 * Contributor(s): Author Derek J Smithies (derek@indranet.co.nz)
 *
 * Note the license of this file was changed to the above with the full
 * permission of the authors.
 *
 * $Log: xlibvid.cxx,v $
 * Revision 1.7  2000/07/06 01:20:18  dereks
 * Free pixel memory on closing the rendering window.
 * Change title.
 *
 * Revision 1.6  2000/07/03 06:35:20  craigs
 * Added ability for video to work on 24, 16 and 8 bit visuals
 *
 * Revision 1.5  2000/05/02 04:32:25  robertj
 * Fixed copyright notice comment.
 *
 */

#ifdef HAS_X11
#include <sys/time.h>
#include <unistd.h>
#include <ptlib.h>
#include "xlibvid.h"

//////////////////////////////////////////////////////////////////////////
//
//  Generic Xlib functions
//

void GenericXlibVideoDevice::SetSize (int _width, int _height)
{
  if ((width != _width) || (height != _height)) {
    CloseWindow();               //Which also closes shared images
    OpenWindow(_width,_height);  //Which also creates Shared Images;
  }
}

/*This routine:
      a)opens a window.
      b)assumes the previous window has been closed.
      c)allocates no memory for rendering of images
      d)calls OpenRender to allocate memory for rendering images.
*/

BOOL GenericXlibVideoDevice::OpenWindow(int _width = 0, int _height = 0)
{
  if ((_width == 0) || (_height == 0))
    return TRUE;

  xLibMutex.Wait();

  // save parameters
  width     = _width; 
  height    = _height;
  n_bytes   = width*height;
  useImage0 = 0;

  // open display and screen
  display = XOpenDisplay(NULL);  
  if (display == NULL) {
    cerr << "Failed to open display in GenericXlibVideoDevice::OpenWindow"<<endl;
    return FALSE;
  }
  screen = DefaultScreen(display);

  // get visual information
  XVisualInfo visualTemplate;
  XVisualInfo * visualInfo;
  Visual * visual;
  int count;
  int index;

  visualTemplate.screen = screen;
  visualInfo = NULL;
  count = index = 0;

  // look for a visual of depth 24
  if ((forceDepth == 0) || (forceDepth == 24)) {
    pixelDepth            = 24;
    byteDepth             = 4;
    visualTemplate.depth  = pixelDepth;
    visualInfo            = XGetVisualInfo(display, VisualScreenMask | VisualDepthMask, &visualTemplate, &count);

    // look for DirectColor and TrueColor visuals
    index = count;
    if (count != 0) 
      for (index = 0; index < count; index++) 
        if (visualInfo[index].c_class == DirectColor ||
            visualInfo[index].c_class == TrueColor)
          break;
  }
  
  // if no matches, try depth 16
  if (((forceDepth == 0) && (index == count)) || (forceDepth == 16)) {
    if (visualInfo != NULL)
      XFree(visualInfo);
    pixelDepth            = 16;
    byteDepth             = 2;
    visualTemplate.depth  = pixelDepth;
    visualInfo            = XGetVisualInfo(display, VisualScreenMask | VisualDepthMask, &visualTemplate, &count);

    // look for DirectColor and TrueColor visuals
    index = count;
    if (count != 0) 
      for (index = 0; index < count; index++) 
        if (visualInfo[index].c_class == DirectColor ||
            visualInfo[index].c_class == TrueColor)
          break;
  }

  // if no matches, try depth 8
  if (((forceDepth == 0) && (index == count)) || (forceDepth == 8)) {
    if (visualInfo != NULL)
      XFree(visualInfo);
    pixelDepth            = 8;
    byteDepth             = 1;
    visualTemplate.depth  = pixelDepth;
    visualInfo            = XGetVisualInfo(display, VisualScreenMask | VisualDepthMask, &visualTemplate, &count);

    // look for PseudoColor visuals
    index = count;
    if (count != 0) 
      for (index = 0; index < count; index++) 
        if (visualInfo[index].c_class == PseudoColor)
          break;
  }

  // if no matches, we are out of luck
  if (index == count) {
    if (visualInfo != NULL)
      XFree(visualInfo);
    if (forceDepth != 0)
      cerr << "Default screen does not support " << forceDepth << " visuals" << endl;
    else
      cerr << "Default screen does not support 24 or 16 bit TrueColor or DirectColor visuals, or 8 bit PseudoColor visuals" << endl;
    return FALSE;
  }

  visual = visualInfo[index].visual;

  // if we ended up with a PseudoColor visual, then use 64 level gray scale
  if (visualInfo[index].c_class != PseudoColor) {
    useGreyScale = FALSE;
    MaskCalc(redPos,   redWidth,   redMask,   visual->red_mask);
    MaskCalc(greenPos, greenWidth, greenMask, visual->green_mask);
    MaskCalc(bluePos,  blueWidth,  blueMask,  visual->blue_mask);
  } else {
    useGreyScale = TRUE;
    int grey = 0;
    Colormap cmap = DefaultColormap(display, screen);
    for (grey = 0; grey < 64; grey++) {
      XColor color;
      color.red = color.green = color.blue = grey << (16 - 5);
      if (!XAllocColor(display, cmap, &color))
        cerr << "Cannot allocate grey scale level " << grey << endl;
      else
        greyColors[grey] = color.pixel;
    }
    pixelDepth            = 8;
    byteDepth             = 1;
  }

  XSetWindowAttributes attributes;
  int attributeMask = 0;

  Colormap colormap = XCreateColormap(display, RootWindow(display, screen), visual, AllocNone);
  attributeMask       |= CWColormap;
  attributes.colormap = colormap;

  attributeMask       |= CWBorderPixel;
  attributes.border_pixel = 0;
  win = XCreateWindow(display,
                      RootWindow(display, screen),
                      10, 10,
                      width, height,
                      5,
                      pixelDepth,
                      InputOutput,
                      visual,       
                      attributeMask, &attributes);
  if (win == 0) {
    cerr << "Allocation of X window failed" << endl;
    return FALSE;
  }

  //XSetWindowColormap(display, win, colormap);

  XSizeHints size_hints;
  size_hints.flags = PSize | PMinSize | PMaxSize;  //Prevent resizing.
  size_hints.min_width = width;
  size_hints.max_width = width;
  size_hints.min_height = height;
  size_hints.max_height = height;

  // Set name of window, used to display receive or send images.
  PString str1;
  if (remoteName *= "Local")
    str1= remoteName;
  else
    str1= PString("From ") & remoteName;
  XSetStandardProperties(display,
                         win,
                         (const char *)str1,
                         (const char *)remoteName,
                         None,0,0,&size_hints); 

  XSelectInput(display, win, ExposureMask|StructureNotifyMask);
  XMapWindow(display, win);
  XRaiseWindow(display,win);
  XFlush(display);

  gc = XCreateGC(display, win, 0, NULL);

  OpenRender(visual);

  XFlush(display);
  XFree(visualInfo);

  xLibMutex.Signal();
  return TRUE;
}

void GenericXlibVideoDevice::MaskCalc(int & pos, int & wid, int & newMask, int mask)
{
  // look for starting position
  pos = 0;
  while ((mask & (1 << pos)) == 0)
    pos++;

  // save mask
  newMask = mask >> pos;

  // look for width
  wid = 1;
  while ((mask & (1 << (pos + wid))) != 0)
    wid++;

  // create final mask
  newMask = newMask << (8 - wid);

  // create final pos
  pos -= (8 - wid);

  //cout << ::hex << "Mask=" << mask << ", m=" << newMask << ::dec << ",w=" << wid << ",p=" << pos << endl;
}

void GenericXlibVideoDevice::SetPixel(BYTE * dest, int x, int y, BYTE r, BYTE g, BYTE b)
{
  unsigned long * val = (unsigned long *)(dest + (x + (y * width)) * byteDepth);

  if (useGreyScale) {
    int grey = (int)((r * 0.299) + (g * 0.587) + (b * 0.114));
    if (grey > 255)
      grey = 255;
    else if (grey < 0)
      grey = 0;
    *val = greyColors[grey >> (8 - 5)];
    return;
  }

#if 0

  *val = ((r & redMask) << redPos) |
         ((g & greenMask) << greenPos) |
         ((b & blueMask) << bluePos);

#else

  *val = 0;

  if (redPos > 0)
    *val |= (r & redMask) << redPos;
  else
    *val |= (r & redMask) >> -redPos;

  if (greenPos > 0)
    *val |= (g & greenMask) << greenPos;
  else
    *val |= (g & greenMask) >> -greenPos;

  if (bluePos > 0)
    *val |= (b & blueMask) << bluePos;
  else
    *val |= (b & blueMask) >> -bluePos;
#endif
}

void GenericXlibVideoDevice::FillRect(BYTE * dest, int x, int y, int width, int height, BYTE R, BYTE G, BYTE B)
{
  int r, c;
  for (r = 0; r < height; r++)
    for (c = 0; c < width; c++)
      SetPixel(dest, x+c, y+r, R, G, B);
}

void GenericXlibVideoDevice::GenerateImage(BYTE * dest)
{
  static BYTE row1[7][3] = {
    { 204, 204, 204 },   // 80% grey
    { 255, 255,   0 },   // yellow
    {   0, 255, 255 },   // cyan
    {   0, 255,   0 },   // green
    { 255,   0, 255 },   // magenta
    { 255,   0,   0 },   // red
    {   0,   0, 255 },   // blue
  };

  static BYTE row2[7][3] = {
    {   0,   0, 255 },   // blue
    {  19,  19,  19 },   // black
    { 255,   0, 255 },   // magenta
    {  19,  19,  19 },   // black
    {   0, 255, 255 },   // cyan
    {  19,  19,  19 },   // black
    { 204, 204, 204 },   // grey
  };

  static BYTE row3a[4][3] = {
    {   8,  62,  89 },   // I
    { 255, 255, 255 },   // white
    {  58,   0, 126 },   // +Q
    {  19,  19,  19 },   // black
  };

  static BYTE row3b[3][3] = {
    {   0,   0,   0 },   // 3.5
    {  19,  19,  19 },   // 7.5
    {  38,  38,  38 },   // 11.5
  };

  static BYTE row3c[3] = { 19,  19,  19 };

  int row1Height = (int)(0.66 * height);
  int row2Height = (int)((0.75 * height) - row1Height);
  int row3Height = height - row1Height - row2Height;

  int colWidth = width / 7;

  PINDEX i;

  // first row
  for (i = 0; i < 7; i++) 
    FillRect(dest, i*colWidth, 0, colWidth, row1Height, row1[i][0], row1[i][1], row1[i][2]);

  // second row
  for (i = 0; i < 7; i++) 
    FillRect(dest, i*colWidth, row1Height, colWidth, row2Height, row2[i][0], row2[i][1], row2[i][2]);

  // third row
  int colB = 5 * (width / 7);
  colWidth = colB / 4;
  for (i = 0; i < 4; i++) 
    FillRect(dest, i*colWidth,        row1Height + row2Height, colWidth, row3Height, row3a[i][0], row3a[i][1], row3a[i][2]);

  colWidth = (width / 7) / 3;
  for (i = 0; i < 4; i++) 
    FillRect(dest, i*colWidth + colB, row1Height + row2Height, colWidth, row3Height, row3b[i][0], row3b[i][1], row3b[i][2]);

  colWidth = width / 7;
  FillRect(dest, (width * 6) / 7,     row1Height + row2Height, colWidth, row3Height, row3c[0], row3c[1], row3c[2]);

}


void GenericXlibVideoDevice::CloseWindow()
{
  xLibMutex.Wait();
  if ((width != 0) && (height != 0)) {
    if (colormap != DefaultColormap(display, screen))
      XFreeColormap(display, colormap);
    CloseRender();
    XFreeGC(display, gc);
    XDestroyWindow(display,win);
    XFlush(display);
    width = 0;
    height= 0;
    n_bytes=0;
  }
  xLibMutex.Signal();
}

#define LIMIT(x) (BYTE)(((x>0xffffff)?0xff0000:((x<=0xffff)?0:x&0xff0000))>>16)

void GenericXlibVideoDevice::Translate411Input(BYTE *dest, const BYTE *frame)
{	
  const BYTE * yplane  = frame;
  const BYTE * uplane  = frame + n_bytes;
  const BYTE * vplane  = frame + n_bytes + (n_bytes >> 2);
  BYTE r,g,b;

  long Y,Cr,Cb;
  long l,lr2,lg2,lb2;
  long lr,lb,lg;
  int x, y;

  struct timeval t1,t2;
  gettimeofday(&t1,NULL);
 
  //In the YUV frame, there are four Y pixels for every UV pair.
  //Thus, calculate the Cr and Cb values for a UV pair, and then
  //determine the RGB values for each of the four Y Pixels.

#define LIMITY if (Y<0) Y=0
#define PROCESS_YPIXEL          l = 76310 * Y;      \
                                lr = l + lr2;       \
                                lg = l + lg2;       \
                                lb = l + lb2;       \
                                r = LIMIT(lr);      \
                                g = LIMIT(lg);      \
                                b = LIMIT(lb)

  for (y = 0; y < height; y+=2) {
    for (x = 0; x < width; x+=2) {
      Cr = *uplane - 128;   //First, get Cb and Cr.
      Cb = *vplane - 128;
      lr2 = 104635 * Cb;
      lg2 = -25690 * Cr + -53294 * Cb;
      lb2 = 132278 * Cr;

      Y  = *yplane - 16;          
      LIMITY;
      PROCESS_YPIXEL;  

      //*rgb= GET_RGB_INT;
      //rgb ++;

      SetPixel(dest, x + 0, y + 0, r, g, b);

      yplane++;
      Y  = *yplane - 16;          
      LIMITY;
      PROCESS_YPIXEL;  

      //*rgb= GET_RGB_INT;
      //rgb ++;
      SetPixel(dest, x + 1, y + 0, r, g, b);

      yplane++;
      Y  = *(yplane+width-2) - 16;          
      LIMITY;
      PROCESS_YPIXEL;  

      //*(rgb+width-2)= GET_RGB_INT;
      SetPixel(dest, x + 0, y + 1, r, g, b);

      Y  = *(yplane+width-1) - 16;          
      LIMITY;
      PROCESS_YPIXEL;  

      //*(rgb+width-1)= GET_RGB_INT;
      SetPixel(dest, x + 1, y + 1, r, g, b);

      uplane++;
      vplane++;
    }
    yplane+=width;
  }

  int diffrence;
  static int total_diff=0;
  static int framesdone=0;
  gettimeofday(&t2,NULL);
  if(t1.tv_sec<t2.tv_sec)
    diffrence=1000000;
   else
    diffrence=0;
  diffrence+= t2.tv_usec-t1.tv_usec;
  total_diff+=diffrence;
  framesdone++;
  if((framesdone%100)==0)
      cerr<<"average us per frame is "<<total_diff/framesdone<<endl;

  return;
}

BOOL GenericXlibVideoDevice::Redraw (const BYTE *frame)
{
  xLibMutex.Wait();
  if(width&&height) {
     Translate411Input(pixels[useImage0], frame);
     XImage * xi = images[useImage0];
     DisplayImage(xi);
     XFlush(display);
     useImage0 = 1 - useImage0;
  }
  xLibMutex.Signal();
  return TRUE;
}

//////////////////////////////////////////////////////////////////////////
//
//  Shared memory Xlib functions
//

void ShmXlibVideoDevice::OpenRender(Visual * visual)
{
  CreateShmXImage(&images[0], &shminfo0, visual);
  CreateShmXImage(&images[1], &shminfo1, visual);  

  pixels[0] = (BYTE *)shminfo0.shmaddr;
  pixels[1] = (BYTE *)shminfo1.shmaddr;
}

void ShmXlibVideoDevice::CloseRender()
{
   DestroyShmXImage(&images[0], &shminfo0);
   DestroyShmXImage(&images[1], &shminfo1);
   free(pixels[0]);
   free(pixels[1]);
}

void ShmXlibVideoDevice::CreateShmXImage(XImage **image, XShmSegmentInfo *xsi, Visual * visual)
{
  *image = XShmCreateImage(display,
                           visual,
                           pixelDepth,
                           ZPixmap,
                           NULL,
                           xsi,
                           width, height);

  if ((*image) == NULL) {
    cerr << "Failed to create shared memory image" << endl;
    return;
  }

  xsi->shmid = shmget(IPC_PRIVATE,
                      (*image)->bytes_per_line*(*image)->height,
                      IPC_CREAT | 0777 );
  if(xsi->shmid<0) {
    cerr << "Failed to create shared memory area" << endl;
    return;
  }

  xsi->shmaddr = (*image)->data = (char *)shmat(xsi->shmid,0 ,0);
  if (xsi->shmaddr == (char *) -1) {
    cerr << "Failed to assign shared memory area to image->data" << endl;
    return;
  }

  xsi->readOnly = False;
  if(!XShmAttach(display, xsi)) {
    cerr << "Failed to attach shared memory area to X server" << endl;
    return;
  }
}

void ShmXlibVideoDevice::DestroyShmXImage(XImage **image, XShmSegmentInfo *xsi)
{
  struct shmid_ds shared_ds;
  if ((*image) != NULL) {
    XShmDetach(display,xsi);
    XDestroyImage(*image);
    shmdt(xsi->shmaddr);
    shmctl(xsi->shmid,IPC_RMID,&shared_ds);          
  }
}

void ShmXlibVideoDevice::DisplayImage(XImage *xi)
{
  XShmPutImage(display,win,gc,xi,0,0,0,0, xi->width,xi->height,False);
}

//////////////////////////////////////////////////////////////////////////
//
//  Non-shared memory Xlib functions
//

void XlibVideoDevice::OpenRender(Visual * visual)
{
  CreateXImage(&images[0], &pixels[0], visual);
  CreateXImage(&images[1], &pixels[1], visual);
}

void XlibVideoDevice::CloseRender()
{
  XDestroyImage(images[0]);
  XDestroyImage(images[1]);
}

void XlibVideoDevice::CreateXImage(XImage **image, BYTE ** pixelData, Visual * visual)
{
  *pixelData = (BYTE *)malloc(width * height * byteDepth);

  if ((*pixelData) == NULL) {
    cerr << "Failed to allocate storage for pixel data" << endl;
    return;
  }
  *image = XCreateImage(display,
                        visual,
                        pixelDepth,
                        ZPixmap,
                        0,
                        (char *)*pixelData,
                        width, height,
                        8, 0);

  if ((*image) == NULL) {
    cerr << "Failed to create X image" << endl;
    return;
  }
}

void XlibVideoDevice::DestroyXImage(XImage **image)
{
  if (*image != NULL)
    XDestroyImage(*image);
}

void XlibVideoDevice::DisplayImage(XImage *xi)
{
  XPutImage(display,win,gc,xi,0,0,0,0, xi->width,xi->height);
}

#endif

